''' Wrap C extension for libpll library for use in natural Python. '''

# Date:   Feb 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import alignment, rearrangement
from libpllWrapper import *
from tempfile import NamedTemporaryFile as NTempFile
import os, sys

class dataModel:
    
    ''' Encapsulating a phylogenetic tree (as topology) + corresponding alignment
    into a libpll-associated data structure. Allows for log-likelihood
    scoring of this model. MUST BE CLOSED AFTER USE. '''

    def __init__(self,topo,alignm,model=None):

        ''' Initialize the data model and respective structures. 
        
        :param topo: a topology object
        :type topo: :class:`.rearrangement.topology`
        :param alignm: a phylip-friendly alignment object.
        :type alignm: :class:`.alignment.phylipFriendlyAlignment`
        
        '''
        
        # Ensure that given alignment object is of correct type.
        if type(alignm) != alignment.phylipFriendlyAlignment:
            raise TypeError('libpll interfacing needs phylipFriendlyAlignment.')

        # Model handling. See if one already defined in alignment.
        if not model:
            if hasattr(alignm,'_pllmodel'): self.model = alignm._pllmodel
            else:
                self.model = partitionModel(alignm)
                self.model.createSimpleModel()
                alignm._pllmodel = self.model
                alignm.paths['pll'] = self.model.getFileName()
        else: self.model = model        

        # Other things.
        modf           = self.model.getFileName()
        self.instance  = new(alignm.getPhylip(),
                             topo.toUnrootedNewick(),
                             modf)
        
    def getNewickString(self):
        
        ''' Acquire the Newick string of the problem instance.
        
        :return: a Newick string
        
        '''
        
        return getNewickString(self.instance)
        
    def getLogLikelihood(self):
        
        ''' Calculates log-likelihood using libpll.
        
        :return: a floating point value
        
        '''
       
        return getLogLikelihood(self.instance)

    def close(self):

        ''' If done with this particular problem. Frees associated memory. '''
        
        destroy(self.instance)

class partitionModel:
    
    ''' A partition model intended for libpll. '''
    
    def __init__(self,ali):
        
        ''' Initialize a partition model (for internal use by libpll). 

        :param ali: an alignment object
        :type ali: :class:`.alignment.alignment`

        '''
        
        self.handle = NTempFile(delete=False)
        self.length = len(ali)
        self.isprotein = (ali.getDataType() == 'protein')
        
    def getFileName(self):
        
        ''' Get the file name of the model file.
        
        :return: a string
        
        '''
        
        return self.handle.name
    
    def createSimpleModel(self,pmodel='WAG'):
        
        ''' Establish a simple model (e.g., one type).
        
        :param pmodel: optional; what protein model to use (as described in pll)
        :type pmodel: a string (default 'WAG')
        :return: None
        
        '''
        
        if self.isprotein: simplemodel = "%s, p1 = 1-%d\n" % (
            pmodel,self.length)
        else: simplemodel = "DNA, p1 = 1-%d\n" % (self.length)
        self.handle.write(simplemodel)
        self.handle.close()
        
    def createModel(self,models,partnames,ranges):
        
        ''' Establish a more complex model.
        
        :param models: a list of model names (e.g., 'WAG', 'DNA')
        :type models: a list of strings
        :param partnames: a list of partition names (e.g., 'p1', 'p2')
        :type partnames: a list of strings
        :param ranges: a list of range tuples (what ranges of alignment)
        :type ranges: a list of integer tuples
        :return: None
        
        '''
        
        for m in xrange(len(models)):
            mod    = models[m]
            par    = partnames[m]
            rl, ru = ranges[m]
            towr   = "%s, %s = %d-%d\n" % (mod,par,rl,ru)
            self.handle.write(towr)
        self.handle.close()
        
    def close(self):
        
        ''' Delete file. '''
        
        os.unlink(self.handle.name)



''' Pylogeny is a Python library and code framework for phylogenetic tree
reconstruction and scoring.

Allows one to perform the following tasks: (1) Generate and manage phylogenetic
tree landscapes. (2) Build and rearrange phylogenetic trees using preset
operators such as NNI, SPR, and TBR. (3) Score phylogenetic trees by
Log-likelihood and Parsimony.

Dependencies: Pandas, P4 Phylogenetic Library. Suggested: FastTree, RAxML,
PytBEAGLEhon. '''

import os as _

thisdir = _.path.split(_.path.realpath(__file__))[0]
itlist = _.listdir(thisdir)
__all__ = [_.path.split(x)[-1].strip('.py') for x in itlist if x.endswith('.py')
           and not x.endswith('init__.py')]


''' Encapsulate a phylogenetic tree space. A phylogenetic landscape or tree
space refers to the entire combinatorial space comprising all possible
phylogenetic tree topologies for a set of n taxa. The landscape of n taxa can be
defined as consisting of a finite set T of tree topologies. Tree topologies can
be associated with a fitness function f(t_i) describing their fit. This forms a
discrete solution search space and finite graph (T, E) = G. E(G) refers to the
neighborhood relation on T(G). Edges in this graph are bidirectional and
represent transformation from one tree topology to another by a tree
rearrangement operator. An edge between t_i and t_j would be notated as e_{ij}
in E(G). '''

# Date:   Jan 24 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# Imports

import networkx
import tree
import alignment
import base
from random import choice
from scoring import getParsimonyFromProfiles as parsimony, getLogLikelihood as\
     ll
from parsimony import profile_set as profiles
from networkx import components as comp, algorithms as alg
from base import patriciaTree
from tree import treeSet, numberRootedTrees, numberUnrootedTrees
from newick import newickParser, removeBranchLengths
from rearrangement import TYPE_NNI, TYPE_SPR, TYPE_TBR
postOrderTraversal = base.treeStructure.postOrderTraversal

LS_NOT_DEFINED = -1

# Graph Object

class graph(object):
    
    ''' Define an empty graph object. '''
    
    def __init__(self,gr=None,defWeight=0.):
        
        ''' Instantiate a graph. Default edge weights are 0.
        
        :param gr: a networkx graph object, if already exists. 
        :type gr: a :class:`networkx.Graph` object
        :param defWeight: the default edge weight of weights
        :type defWeight: a floating point number
        
        '''
        
        if gr == None: self.graph = networkx.Graph()
        else: self.graph = gr
        self.defaultWeight = defWeight
    
    def __len__(self):
        
        return len(self.graph.node)
    
    def __iter__(self):
        
        for node in self.graph.node.keys(): yield node

    def getSize(self):

        ''' Return the number of nodes in the graph.
        
        :return: an integer
        
        '''
        
        return len(self.graph.node)

    def getNodeNames(self):
        
        ''' Return the names of nodes in the graph.
        
        :return: a list of strings
        
        '''
        
        return self.graph.node.keys()
    
    def iterNodes(self):
        
        ''' Iterate over all node keys.
        
        :return: a generator of node keys
        
        '''
        
        for node in self.graph.node: yield node
        
    def getNodes(self):
        
        ''' Get all node values. Recommended to use an iterator for large
        graphs (iterNodes()).
        
        :return: a list of node values (whatever is associated with nodes)
        
        '''
        
        return self.graph.node.values()
    
    def getEdges(self):
        
        ''' Get all edges (as defined for NetworkX graphs). Recommended to use
        an iterator for large graphs.
        
        :return: a list of edges (and associated data)
        
        '''
        
        return [self.getEdge(i,j) for i,j in self.graph.edges_iter()]
    
    def getEdgesFor(self,i):
        
        ''' Get all edges associated with a certain node. 
        
        :param i: a node name
        :type i: a string
        :return: a list of edges (and associated data) for all neighbors
        
        '''
        
        return [self.getEdge(i,j) for j in self.graph.neighbors(i)]
    
    def getNode(self,i):
        
        ''' Get a single node by name. 
        
        :param i: a node name:
        :type i: a string
        :return: a dictionary (with node information)
        
        '''
        
        return self.graph.node[i] 
    
    def getEdge(self,i,j):
        
        ''' Get the data associated with an edge (including weight).
        
        :param i: a node name
        :type i: a string
        :param j: a node name
        :type j: a string
        :return: an edge (and associated data)
        
        '''
        
        return self.graph.get_edge_data(i,j)
    
    def isEdge(self,i,j):
        
        ''' See if an edge exists between two nodes.
        
        :param i: a node name
        :type i: a string
        :param j: a node name
        :type j: a string
        :return: a boolean
        
        '''
        
        return (self.getEdge(i,j) != None)
    
    def getNeighborsFor(self,i):
        
        ''' Get a list of all node names neighbor to a node. 
        
        :param i: a node name
        :type i: a string
        :return: a list of strings (node names)
        
        '''
        
        return self.graph.neighbors(i)  
    
    def getDegreeFor(self,i):

        ''' Return in- and out-degree for node named i.
        
        :param i: a node name
        :type i: a string
        :return: an integer
        
        '''
        
        return len(self.getNeighborsFor(i))
        
    def setDefaultWeight(self,w):
        
        ''' Set the default weight of edges (weight of edges if not overridden).
        
        :param w: a weight
        :type w: a floating point
        
        '''
        
        self.defaultWeight = float(w)
        
    def clearEdgeWeights(self):
        
        ''' Set all edge weights to the default edge weight. '''
        
        for edge in self.getEdges():
            edge['weight'] = self.defaultWeight
        
    def getNumComponents(self):
        
        ''' Get the number of components of the graph.
        
        :return: an integer
        
        '''
        
        return comp.number_connected_components(self.graph)
    
    def getComponents(self):
        
        ''' Get the connected components in the graph. '''
        
        return comp.connected_components(self.graph)
        
    def getComponentOfNode(self,i):
        
        ''' Get the graph component of a given node. '''
        
        return comp.node_connected_component(self.graph,i)

    def getCliques(self):
        
        ''' Get the cliques present in the graph. '''
        
        return alg.clique.find_cliques(self.graph)
    
    def getCliqueNumber(self):

        ''' Get the clique number of the graph.
        
        :return: an integer
        
        '''
        
        return alg.clique.graph_clique_number(self.graph)
    
    def getNumCliques(self):
        
        ''' Get the number of cliques found in the graph.
        
        :return: an integer
        
        '''
        
        return alg.clique.number_of_cliques(self.graph)
    
    def getCliquesOfNode(self,i):
        
        ''' Get the clique that a node corresponds to. '''
        
        return alg.clique.cliques_containing_node(self.graph,i)
    
    def getCenter(self):
        
        ''' Get the centre of the graph. '''
        
        return alg.distance_measures.center(self.graph)
    
    def getDiameter(self):
        
        ''' Acquire the diameter of the graph. '''
        
        return alg.distance_measures.diameter(self.graph)    
    
    def getMST(self):

        ''' Acquire the minimum spanning tree for the graph. '''
        
        return alg.minimum_spanning_tree(self.graph)
    
    def hasPath(self,nodA,nodB):

        ''' See if a path exists between two nodes. '''
        
        return alg.has_path(self.graph,nodA,nodB)

    def getShortestPath(self,nodA,nodB):
        
        ''' Get the shortest path between two nodes. '''
        
        return alg.shortest_path(self.graph,nodA,nodB)

    def getShortestPathLength(self,nodA,nodB):
        
        ''' Get the shortest path length between two nodes. '''
        
        return alg.shortest_path_length(self.graph,nodA,nodB)


class landscape(graph,treeSet):
    
    ''' Defines an entire phylogenetic tree space. '''
    
    def __init__(self,ali,starting_tree=None,root=True,operator='SPR'):
        
        ''' Initialize the landscape.

        :param ali: an alignment
        :type ali: an :class:`.alignment.alignment` object
        :param starting_tree: an optional tree object to start with
        :type starting_tree: a :class:`.tree.tree` object
        :param root: whether or not to compute an approximate\
        maximum likelihood tree (FastTree) or start the landscape\
        with a given starting tree.
        :type root: a boolean
        :param operator: a string that describes what operator the\
        landscape is mostly comprised of.
        :type operator: a string
        
        '''        
        
        super(landscape,self).__init__()
        
        # Fields
        self.alignment          = ali
        self.locks              = list()
        self.nextTree           = 0
        self.leaves             = None
        self.root               = None
        self.operator           = operator
        self.parsimony_profiles = None
        self.newickSearchDict   = dict()
        
        # Analyze alignment.
        if ali:
            # Get number of leaves.
            self.leaves = ali.getNumSeqs()
            # Get parsimony profile.
            self.parsimony_profiles = profiles(ali)           
                
        # Set up the root.
        if root:
            if starting_tree == None:
                approx      = ali.getApproxMLTree()
                self.root   = self.addTree(approx)
            else: self.root = self.addTree(starting_tree)
            # Score this tree.
            node  = self.getNode(self.root)
            tre   = node['tree']
            topol = tre.toTopology()
            new   = topol.toNewick()
            if not (tre.score):
                sc = parsimony(new,p)
                tre.score = (None,sc)            
        
    def getAlignment(self): 
        
        ''' Acquire the alignment object associated with this space. 
        
        :return: an :class:`.alignment.alignment` object
        
        '''
        
        return self.alignment
    
    def getNumberTaxa(self):
        
        ''' Return the number of different taxa present in any respective tree
        in the landscape. 
        
        :return: an integer
        
        '''
        
        return self.leaves

    def getPossibleNumberRootedTrees(self):
        
        ''' Assuming all of the trees in the space are rooted, return the
        maximum possible number of unrooted trees that can possibly be generated
        for the number of taxa of trees in the landscape.
        
        :return: an integer
        
        '''
        
        return numberRootedTrees(self.leaves)
    
    def getPossibleNumberUnrootedTrees(self):
        
        ''' Assuming all of the trees in the space are unrooted, return the 
        maximum possible number of unrooted trees that can possibly be generated
        for the number of taxa of trees in the landscape.
        
        :return: an integer
        
        '''
        
        return numberUnrootedTrees(self.leaves)
    
    def getRoot(self):
        
        ''' Returns the index to the root (starting) tree of the space.
        
        :return: an integer
        
        '''
        
        return self.root
    
    def getRootNode(self):
        
        ''' Returns the root (starting) tree of the space in its node form. 
        
        :return: a dictionary (with node information)
        
        '''
    
    def getRootTree(self):
        
        ''' Acquire the first tree that was placed in this space.
        
        :return: a :class:`.tree.tree` object
        
        '''
        
        return self.getTree(self.root)
    
    def setAlignment(self,ali):
        
        ''' Set the alignment present in this landscape. WARNING; will not
        modify existing scores. 
        
        :param ali: an alignment
        :type ali: an :class:`.alignment.alignment` object
        
        '''
        
        self.alignment          = ali
        self.leaves             = ali.getNumSeqs()
        self.parsimony_profiles = profiles(ali)
        
    def setOperator(self,op):
        
        ''' Set the operator assigned to this landscape.
        
        :param op: an operator (string description)
        :type op: a string
        
        '''
        
        self.operator = op
    
    # Node Management
                
    def _newNode(self,tobj,score=False):
        
        ''' PRIVATE: Add a new node. '''
        
        # Extract data from the object.
        if type(tobj) != tree.tree:
            raise TypeError(
                'Nodes in space must be constructed from tree objects.')
        name      = tobj.getName()
        structure = tobj.getStructure()
        
        # Add its structure to a dictionary structure.
        query = self.findTreeTopologyByStructure(structure)
        if query != None:
            raise AssertionError('Tree (%s) <%s> already exists in space!' % (
                str(name),str(structure)))
        i = len(self) # Get next possible value for insertion unique integer.
        self.newickSearchDict[structure] = i
        
        # Create the node.
        self.graph.add_node(i)
        node = self.graph.node[i]
        node['index']    = i
        node['explored'] = False
        node['tree']     = tobj
        node['failed']   = False
        
        # Preliminary scoring.
        if score:
            if tobj.score == None:
                # Get the parsimony since this is fast; set likelihood
                # to nothing.
                if self.alignment:
                    tobj.score = (None,parsimony(
                        tobj.newick,self.parsimony_profiles))
                else: tobj.score = (None,None)
            elif tobj.score[1] == None:
                # Get the parsimony since this is fast; set likelihood
                # to nothing.
                if self.alignment:
                    tobj.score = (tobj.score[0],parsimony(
                        tobj.newick,self.parsimony_profiles))  
        
        # Return the index.
        return i
    
    def getTree(self,i):
        
        ''' Get the object for a tree by its name.
        
        :param i: a tree name (usually an integer)
        :return: a :class:`.tree.tree` object
        
        '''
        
        if not i in self.graph.node: return None
        return self.getNode(i)['tree']
    
    def __iter__(self):
        
        for t in self.graph.nodes():
            yield self.getTree(t)

    def getVertex(self,i):
        
        ''' Acquire a vertex object from the landscape; this is a 
        high-level representation of a tree in the landscape with
        additional functionality. Object created upon invocation of
        this function.
        
        :param i: a tree name (usually an integer)
        :return: a :class:`.vertex` object
        
        '''
        
        return vertex(self.getNode(i),self)

    def removeTreeByIndex(self,i):
        
        ''' Remove a tree from the landscape by index.
        
        :param i: a tree name (usually an integer)
        :return: a boolean (success or failure)
        
        '''
        
        tr = self.getTree(i)
        if (tr == None): return False
        self.graph.remove_node(i)
        del self.newickSearchDict[tr.getStructure()]
        return True

    def removeTree(self,tree):
        
        ''' Remove a tree from the landscape by object.
        
        :param tree: a tree that exists in the landscape
        :type tree: a :class:`.tree.tree` object
        :return: a boolean (success or failure)
        
        '''

        t = self.indexOf(tree)
        if (t < 0): return False
        return self.removeTreeByIndex(t)
        
    def addTreeByNewick(self,newick,score=True,check=True,struct=None):
        
        ''' Add tree to the landscape by Newick string. Will return index.
        
        :param newick: a Newick string
        :type newick: a string
        :param score: defaults to True, whether to score this tree or not
        :type score: a boolean
        :return: the index of the tree
        
        '''
        
        return self.addTree(None,score=score,check=check,newick=newick,
                            struct=struct)
        
    def addTree(self,tr,score=True,check=True,newick=None,struct=None):
        
        ''' Add a tree to the landscape. Will return its index. 

        :param tr: a tree
        :type tr: a :class:`.tree.tree` object
        :param score: defaults to True, whether to score this tree or not
        :type score: a boolean
        :return: the index of the tree
        
        '''
            
        # (Re)construct the tree ensuring consistent configuration/naming.
        if tr == None and newick == None:
            raise ValueError(
                'Require tree object OR Newick string for tree addition.')
        if newick == None: newick = tr.toNewick()
        tobj = tree.tree(newick,check=check,structure=struct)
        
        # See if needs to be scored.
        if not score and tr != None: tobj.setScore(tr.getScore())
        
        # Origin needs to be defined?
        if tr != None: tobj.setOrigin(tr.getOrigin())
        
        # Add to graph.
        index = self._newNode(tobj,score=score)
        if (self.root is None):
            if (len(self) != 1):
                raise AssertionError('Landscape has no root with 1+ tree.')
            self.root = index
        return index
    
    def exploreRandomTree(self,i,type=TYPE_SPR):
        
        ''' Acquire a single neighbor to a tree in the landscape by performing a
        random rearrangement of type SPR (by default), NNI, or TBR -- this is
        done by performing a rearrangement on a random branch in the topology.
        Rearrangement type is provided as a rearrangement module type definition
        of form, for example, TYPE_SPR, TYPE_NNI, etc.
        
        :param i: a tree index
        :param type: the type of rearrangement (e.g., TYPE_SPR, TYPE_NNI)
        :return: the new tree index or None in case of failure
        
        '''
        
        # Get parsimony profiles.
        p = self.parsimony_profiles
        
        # Check node.
        node  = self.getNode(i)
        if (node['explored']): return None
        tre   = node['tree']
        topol = tre.toTopology()
        new   = topol.toNewick()
        isvi  = self._isViolating(topol)
        
        # Check for locks.
        if hasattr(self,'locks'):
            for lock in self.locks:
                hasthis = topol.getBranchFromBipartition(lock)
                if (hasthis): topol.lockBranch(hasthis)
        
        # Perform an exploration (not yet done).
        branches = topol.getBranches()
        while (len(branches) > 0): # Keep performing until unique rearrangement.
            
            bra  = choice(branches)
            branches.remove(bra)
            enum = topol.iterTypeForBranch(bra,type) # Iterate over each.
        
            for en in enum:
                
                # Get metadata.
                typ = en.getType()
                t   = en.toTree()
                new = t.getNewick()
    
                # See if already been found.
                inlandscape = self.findTreeTopologyByStructure(t.getStructure())
                if (inlandscape != None):
                    # Is already in landscape; has connection to tree?
                    if (inlandscape != i and self.graph.has_node(inlandscape)):
                        if not self.graph.has_edge(inlandscape,i):
                            self.graph.add_edge(inlandscape,i)
                            self.getEdge(inlandscape,i)['weight'] = \
                                self.defaultWeight
                    continue
    
                # See if tree violating existing locks.
                if isvi:
                    en = en.toTopology()
                    if self._isViolating(en): continue
    
                # Score.
                scr = parsimony(new,p)
                t.score  = (None,scr)
                t.origin = typ
                
                # Add to landscape.
                j = self._newNode(t,score=True)
                self.graph.add_edge(i,j)
                self.getEdge(i,j)['weight'] = self.defaultWeight
                return j
        
        # Set explored to True.
        node['explored'] = True 
        
        return None        

    def exploreTree(self,i,type=TYPE_SPR):
        
        ''' Get all neighbors to a tree named i in the landscape using a
        respective rearrangement operator as defined in the rearrangement
        module. Rearrangement type is provided as a rearrangement module type
        definition of form, for example, TYPE_SPR, TYPE_NNI, etc. By default,
        this is TYPE_SPR.
        
        :param i: a tree index
        :param type: the type of rearrangement (e.g., TYPE_SPR, TYPE_NNI)
        :return: a list of neighbors as tree names (usually integers)
        
        '''
        
        # Get parsimony profiles.
        p = self.parsimony_profiles
        
        # Check node.
        node  = self.getNode(i)
        if (node['explored']): return list()
        tre   = node['tree']
        topol = tre.toTopology()
        new   = topol.toNewick()
        isvi  = self._isViolating(topol)
        
        # Check for locks.
        if hasattr(self,'locks'):
            for lock in self.locks:
                hasthis = topol.getBranchFromBipartition(lock)
                if (hasthis): topol.lockBranch(hasthis)
        
        # Perform full-enumeration exploration (1 move).
        neighbors = list()        
        enum = topol.allType(type)
        
        for en in enum:
            
            # Get metadata.
            typ = en.getType()
            t   = en.toTree()
            new = t.getNewick()

            # See if already been found.
            inlandscape = self.findTreeTopologyByStructure(t.getStructure())
            if (inlandscape != None):
                # Is in landscape; has connection to tree?
                if ((inlandscape != i) and self.graph.has_node(inlandscape)): 
                    if not self.graph.has_edge(inlandscape,i):
                        self.graph.add_edge(inlandscape,i)
                        self.getEdge(inlandscape,i)['weight'] = \
                            self.defaultWeight
                continue

            # See if tree violating existing locks.
            if isvi:
                en = en.toTopology()
                if self._isViolating(en): continue

            # Score.
            scr = parsimony(new,p)
            t.score  = (None,scr)
            t.origin = typ
            
            # Add to landscape.
            j = self._newNode(t,score=True)
            neighbors.append(j)
            self.graph.add_edge(i,j)
            self.getEdge(i,j)['weight'] = self.defaultWeight
        
        # Set explored to True.
        node['explored'] = True 
        
        return neighbors

    # Lock Management
        
    def getLocks(self):

        ''' Get all restrictions (locks which cannot be violated on splits).
        
        :return: a list of :class:`.tree.bipartition` objects
        
        '''
        
        return self.locks 
    
    def toggleLock(self,lock):
        
        ''' Add a biparition to the list of locked bipartitions if not 
        present; otherwise, remove it. Return status of lock.
        
        :param lock: a bipartition that cannot be violated
        :type lock: a :class:`.tree.bipartition` object
        :return: a boolean (on or off)
        
        '''
        
        toggle = False
        if lock in self.getLocks():
            self.locks.remove(lock)
        else:
            self.locks.append(lock)
            toggle = True
        return toggle
    
    def lockBranchFoundInTree(self,tr,br):
        
        ''' Given a tree node and a branch object, add a given 
        bipartition to the bipartition lock list. Returns 
        bipartition if locked.
        
        :param tr: a tree name
        :param br: a branch in that tree
        :type br: a :class:`.base.treeBranch` object
        :return: a :class:`.tree.bipartition` object that has been locked or None
        
        '''

        # Check for presence of tree.
        g = self.graph.node
        if (not tr in g): return None
        
        # Acquire metadata for tree.
        tr = self.getTree(tr)
        topl = tr.toTopology()
        
        # Get bipartition for branch.
        bipa  = tree.bipartition(topl,br)
        
        # Toggle the lock.
        self.toggleLock(bipa)
        return bipa
    
    def getBipartitionFoundInTreeByIndex(self,tr,brind,topol=None):
        
        ''' Given a tree node and a branch index, return the associated
        bipartition.
        
        :param tr: a tree name
        :param brind: a branch index in that tree, a la post-order traversal
        :return: a :class:`.tree.bipartition` object
        
        '''
        
        # Check for presence of tree.
        if (not topol) and (not tr in self.graph.node): return None
        
        # Acquire metadata for tree.
        if not topol:
            tr = self.graph.node[tr]['tree']
            topl = tr.toTopology()
        
        # Get bipartition for branch.
        nodes = postOrderTraversal(topl.getRoot())
        if not brind in xrange(len(nodes)): return None
        bnode = nodes[brind]
        bipa  = tree.bipartition(topl,bnode.parent)
        
        # Return the bipartition.
        return bipa
            
    def lockBranchFoundInTreeByIndex(self,tr,brind):
        
        ''' Given a tree node and a branch index, add an associated 
        bipartition to the bipartition lock list. Returns the
        bipartition if locked.
        
        :param tr: a tree name
        :param brind: a branch index
        :return: a :class:`.tree.bipartition` object if it has been locked
        
        '''

        # Check for presence of tree.
        if (not tr in self.graph.node): return None
        
        # Acquire metadata for tree.
        tree = self.graph.node[tr]['tree']
        topl = tree.toTopology()
        
        # Get bipartition for branch.
        bipa = self.getBipartitionFoundInTreeByIndex(tr,brind,topl)
        
        # Toggle the lock.
        self.toggleLock(bipa)
        return bipa

    def _isViolating(self,topo):
        
        ''' PRIVATE: Given a topology, determine if it is
        violating existing locks. '''
        
        if len(self.locks) > 0:
            for lock in self.locks:
                hasit = topo.getBranchFromBipartition(lock)
                if (not hasit): return True
        return False
    
    def isViolating(self,i):
        
        ''' Determine if a tree is violating any locks intrinsic 
        to the landscape. Will also return False if the tree (name) is not
        present in the landscape.
        
        :param i: a tree name (usually an integer)
        
        '''
    
        if i in self.graph.node.keys():
            node = self.getNode(i)
            topo = node['tree'].toTopology()
            return self._isViolating(topo)
        return False

    # Search Methods

    def __getitem__(self, i):
        
        return self.graph.node[i]

    def indexOf(self, tr):
        
        ''' Acquire the index/name in this landscape of a tree object. Returns
        -1 if not found. Warning: naively performs a sequential search.
        
        :param tr: a tree
        :type tr: a :class:`.tree.tree` object
        :return: a tree name (usually an integer index) or -1 if not found
        
        '''
        
        for t in self.graph.node:
            if self.getTree(t) == tr: return t
        return -1

    def findTree(self,newick):
        
        ''' Find a tree by Newick string, taking into account branch lengths.
        Returns the index of this tree in the landscape. Warning: naively
        performs a sequential search.
        
        :param newick: a Newick string
        :type newick: a string
        :return: a tree name (usually an integer index) or None if not found
        
        '''
        
        for t in self.graph.node:
            if self.getTree(t).newick == newick: return t
        return None
    
    def findTreeTopology(self,newick):
        
        ''' Find a tree by topology, not taking into account branch lengths.
        
        :param newick: a Newick string
        :type newick: a string
        :return: a tree name (usually an integer index) or None if not found
        
        '''
        
        s = newickParser(newick).parse()
        removeBranchLengths(s)
        s = str(s) + ';'
        return self.findTreeTopologyByStructure(s)
    
    def findTreeTopologyByStructure(self,struct):
        
        ''' Find a tree by topology, not taking into account branch lengths,
        given the topology.
        
        :param struct: a Newick string without branch lengths (a "structure")
        :type struct: a string
        :return: a tree name (usually an integer index) or None if not found
        
        '''
        
        query = (struct in self.newickSearchDict)
        if query is True:
            index = self.newickSearchDict[struct]
            return index
        return None
        
    def getBestImprovement(self,i):
        
        ''' For a tree in the landscape, investigate neighbors to find 
        a tree that leads to the best improvement of fitness function score
        on the basis of likelihood.
        
        :param i: a tree name (usually an integer)
        :return: a tree name (usually an integer) or None if no better tree
        
        '''
        
        nodes = self.getNode
        tree  = self.getTree
        ml    = lambda d: tree(d).getScore()[0]
        if (not i in self.getNodeNames()):
            raise LookupError('No tree by that name in landscape.' % (i))
        node  = nodes(i)
        near  = self.graph.neighbors(i)
        if (len(near) == 0): return None
        best  = max(near,key=ml)
        if (best != None and ml(best) > ml(i)):
            return best
        else: return None
        
    def getPathOfBestImprovement(self,i):
        
        ''' For a tree in the landscape, investigate neighbors iteratively until
        a best path of score improvement is found on basis of likelihood.
        
        :param i: a tree name (usually an integer)
        :return: a list of tree names (usually integers)
        
        '''
        
        path  = list()
        node = self.getNode(i)
        impr = self.getBestImprovement(i)
        curs = node
        if impr == None: return path
        
        while (impr != None):
            path.append(impr)
            impr = self.getBestImprovement(impr)
        
        return path
    
    def getAllPathsOfBestImprovement(self):
        
        ''' Return all paths of best improvement as a dictionary.
        
        :return: a dictionary of tree name to paths (lists of tree names)
        
        '''
        
        paths = dict()
        for node in self.graph.node:
            paths[node] = self.getPathOfBestImprovement(node)
        return paths    
    
    def iterAllPathsOfBestImprovement(self):
        
        ''' Return an iterator for all paths of best improvement.
        
        :return: a generator of paths (lists of tree names)
        
        '''
        
        for node in self.graph.node:
            yield self.getPathOfBestImprovement(node)
    
    def isLocalOptimum(self,i):
        
        ''' Determine if a tree is a local optimum. This means it has the
        following properties: 
        
          (1) Possesses a likelihood score.
          (2) Local neighborhood completely enumerated (and scored).
          (3) None of its neighbors is a better improvement.
        
        :return: a boolean
        
        '''
        
        nodes = self.graph.node
        if (self.getTree(i).score[0] == None):
            return False
        elif (not self.getNode(i)['explored']):
            return False
        p = self.getPathOfBestImprovement(i)
        if (len(p) == 0):
            near = self.graph.neighbors(i)
            if (len(near) == 0): return False
            for node in [nodes[x] for x in near]:
                if node['tree'].score[0] == None and \
                   'failed' in node and not node['failed']:
                    return False
            return True
        return False
    
    def getLocalOptima(self):
        
        ''' Get all trees in the landscape that can be labelled as a local
        optimum.
        
        :return: a list of tree names (usually integers)
        
        '''
        
        opt = []
        for node in self.graph.node:
            if (self.isLocalOptimum(node)): opt.append(node)
        return opt
    
    def getGlobalOptimum(self):
        
        ''' Get the global optimum of the current space.
        
        :return: a tree name (usually an integer)
        
        '''
        
        # TODO: store global optimum as trees are added?
        optima = self.getLocalOptima()
        return max(optima,key=lambda d: self.getTree(d).score[0])

    # Output Methods
    
    def __str__(self): return self._dump()
    
    def _dump(self):
        
        trees = []
        for t in self.getNodeNames(): trees.append(self.getVertex(t))
        if type(self.alignment) == alignment.phylipFriendlyAlignment:
            return '\n'.join([t.getProperNewick() for t in trees])
        else: return '\n'.join([t.getNewick() for t in trees])

    def _toTreeSet(self, func):
        
        treeset = treeSet()
        for t in self.getNodeNames():
            treeset.addTree(func(t))
        return treeset

    def toProperNewickTreeSet(self):
        
        ''' Convert this landscape into an unorganized set of trees where taxa
        names are transformed to their original form (i.e. not transformed to a
        state friendly for the Phylip format).
        
        :return: a :class:`.tree.treeSet` object
        
        '''
        
        return self._toTreeSet(
            lambda t: tree(self.getVertex(t).getProperNewick()))
    
    def toTreeSet(self):

        ''' Convert this landscape into an unorganized set of trees.
        
        :return: a :class:`.tree.treeSet` object
        
        '''
        
        return self._toTreeSet(lambda t: self.getTree(t))

# Comprising Vertices of Landscapes


class vertex(object):
    
    ''' Encapsulate a single vertex in the landscape and add convenient
    functionality to alias parent landscape functions. '''
    
    def __init__(self, obj, ls):
        
        ''' Initialize this vertex. '''
        
        self.id  = obj['index']
        self.obj = obj
        self.ls = ls
        self.ali = ls.alignment
    
    def getIndex(self):
        
        ''' Get the index of this tree in the space.
        
        :return: a tree name (usually an integer)
        
        '''
        
        return self.id
    
    def getDict(self):
        
        ''' Get the dictionary object (key-value pairs) associated with this
        tree as it is in the NetworkX graph. 
        
        :return: a dictionary
        
        '''
        
        return self.obj
    
    def getObject(self):
        
        ''' Get the dictionary object (key-value pairs) associated with this
        tree as it is in the NetworkX graph. 

        :return: a dictionary

        '''
        
        return self.getDict()
    
    def getTree(self):
        
        ''' Get the tree object associated with this tree. 
        
        :return: a :class:`.tree.tree` object
        
        '''
        
        return self.obj['tree']
    
    def getNewick(self):
        
        ''' Get the Newick string of this tree. 
        
        :return: a string
        
        '''
        
        return self.getTree().getNewick()
    
    def getScore(self):
        
        ''' Get (any) score(s) associated with this tree. 
        
        :return: a tuple of floating point values (scores)
        
        '''
        
        return self.getTree().score
    
    def getOrigin(self):
        
        ''' Get the origin of this tree (how it was acquired).
        
        :return: a string
        
        '''
        
        return self.getTree().origin
    
    def getNeighbors(self):
        
        ''' Get any neighbors to this tree in the landscape. 
        
        :return: a list of tree names (usually integers)
        
        '''
        
        return self.ls.graph.neighbors(self.id)
    
    def getDegree(self):
        
        ''' Get the degree of this tree in the graph. 
        
        :return: an integer
        
        '''
        
        return len(self.getNeighbors())
    
    def isLocalOptimum(self):
        
        ''' Determine if this tree is an optimum. 
        
        :return: a boolean
        
        '''
        
        return self.ls.isLocalOptimum(self.id)
    
    def isExplored(self):
        
        ''' See if this tree has had all possible rearrangements performed. 
        
        :return: a boolean
        
        '''
        
        return self.obj['explored']
    
    def isFailed(self):
        
        ''' Determine if any errors are associated with this node.
        
        :return: a boolean
        
        '''
        
        return ('failed' in self.obj and self.obj['failed'])
    
    def setExplored(self, exp):
        
        ''' Override the "explored" flag of this node in the landscape.
        
        :param exp: a boolean
        
        '''
        
        self.obj['explored'] = exp
    
    def approximatePossibleNumNeighbors(self):
        
        ''' Approximate the possible number of neighbors to this vertex
        by considering the type of tree rearrangement operator. Returns
        LS_NOT_DEFINED if the operator is not known yet.
        
        :return: an integer
        
        '''

        n = self.ls.getNumberTaxa()
        if self.ls.operator == 'SPR': return 4*(n-3)*(n-2)
        return LS_NOT_DEFINED
    
    def scoreLikelihood(self):
    
        ''' Acquire the log-likelihood for this vertex.
        
        :return: the log-likelihood score
        
        '''
        
        if self.getScore()[0] == None:
            l = ll(self.getTree(),self.ali)
            self.obj['tree'].score = (l,self.getScore()[1])
            return l
        else: return self.getScore()[0]
    
    def getBestImprovement(self):
        
        ''' Alias function for function of same name in parent
        landscape. '''
        
        return self.ls.getBestImprovement(self.id)
    
    def getPathOfBestImprovement(self):
    
        ''' Alias function for function of same name in parent
        landscape. '''
        
        return self.ls.getPathOfBestImprovement(self.id)
    
    def isBestImprovement(self):
        
        ''' Check to see if this vertex is a best move for another node.
        
        :return: a boolean
        
        '''
        
        for neighbor in self.getNeighbors():
            if self.ls.getTree(neighbor).score[0] < self.getTree().score[0]:
                if self.ls.getBestImprovement(neighbor) == self.id:
                    return True
        return False
    
    def isViolating(self):
        
        ''' Alias function for function of same name in parent
        landscape. '''
        
        return self.ls.isViolating(self.id)

    def getProperNewick(self):
        
        ''' Get the proper Newick string for a tree. 
        :returns: A string.'''
        
        if self.ali == None: return self.getNewick()
        return self.ali.reinterpretNewick(self.getNewick())
    
    def iterBipartitions(self):
        
        ''' Return a generator to iterate over all bipartitions for this vertex.
        '''
    
        # Get tree object.
        tr = self.getTree()
        
        # Get corresponding topology.
        topo  = tr.toTopology()
        nodes = postOrderTraversal(topo.getRoot())
        
        # Get branches + bipartitions of the topology.
        bp = []
        br = [x.parent for x in nodes]
        for b in br:
            if b:
                bi = tree.bipartition(topo,b)
                yield bi
            else: yield None
    
    def getBipartitions(self):
    
        ''' Get all bipartitions for this vertex.
        
        :return: a list of :class:`.tree.bipartition` objects
        
        '''
        
        # Return the bipartitions.
        return [bipart for bipart in self.iterBipartitions()]
    
    def getBipartitionScores(self):
        
        ''' Get all corresponding bipartition vectors of SPR scores. '''
        
        return [x.getSPRScores() for x in self.getBipartitions()]
    
    def getNeighborsOfBipartition(self, bi):
        
        ''' Get corresponding neighbors of a bipartition in this vertex's tree.
        '''
        
        l     = self.ls
        re    = bi.getSPRRearrangements()
        lsids = [l.findTreeTopology(x.toTree().getStructure()) for x in re]
        return lsids
    
    def getNeighborsOfBranch(self, br):
        
        ''' Get corresponding neighbors of a branch in this vertex's tree. '''
        
        tr   = self.getTree()
        topl = tr.toTopology()
        bipa = tree.bipartition(topl,br)
        return self.getNeighborsOfBipartition(bipa)

    



''' Serialize a phylogenetic landscape into a JSON object. '''

# Date:   Jan 27 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from landscapeWriter import landscapeWriter
import json
import landscape

class JSONWriter(landscapeWriter):
    
    ''' Writes a landscape and associated node information to a JSON
    object. '''
    
    def __init__(self, ls, name):
        
        ''' Instantiates this writer.
        
        :param ls: a landscape object
        :type ls: a :class:`.landscape.landscape` object
        :param name: the name of this landscape
        :type name: a string
        
        '''
        
        super(JSONWriter,self).__init__(ls,name)

    def nodeToJSON(self,node):

        ''' Returns a JSON formatted node, given a node ID.
        
        :param node: a name of a tree/node in the graph
        :type node: a string
        
        '''

        n = self.landscape.getVertex(node)
        ml,pars = n.getScore()
        og      = n.getOrigin()
        lmax    = n.isLocalOptimum()
        mimpr   = n.getBestImprovement()
        expl    = n.isExplored()
        dg      = len(n.getNeighbors())
        val     = int(n.isFailed())
        return {'name':str(node),'group':int(node),
               'ml':ml,'parsimony':pars,'origin':og,
               'explored':expl,'degree':dg,
               'isLocalMax':lmax,'maximprov':str(mimpr),
               'value':val} 

    def _jsonNodes(self):
        
        for node in self.graph.node:
            if self.graph.degree(node) > 0:
                yield self.nodeToJSON(node)
    
    def _jsonEdges(self,jsonNodes):
        
        nodeli = [x['group'] for x in jsonNodes]
        for node in nodeli:
            neighbors = self.graph.neighbors(node)
            for neigh in neighbors:
                if neigh in nodeli:
                    yield {'source':nodeli.index(node),
                           'target':nodeli.index(neigh),
                           'value': 1}    
    
    def _jsonGraph(self,nodes=None):
        
        # Initialize variables.
        if not nodes: nodes = self._jsonNodes()
        ls       = self.landscape
        locks    = ls.locks
        toVis    = []
        allScore = True

        # Check scores.
        for node in nodes:

            # Get node's log-likelihood score.
            nS = node['ml']
            if (nS == None): allScore = False
            else: toVis.append(node)

        if len(toVis) == 0: allScore = False        
        return toVis, allScore
    
    def getOnlyImprovements(self,groups=None):
        
        # Are there any groups to uncollapse?
        if (groups == None): groups = list()
        
        # Get all relevant nodes.
        nodes,inds = [self.nodeToJSON(0)],[0]
        paths = self.landscape.iterAllPathsOfBestImprovement()
        for path in paths:
            for it in path:
                if not it in inds:
                    nodes.append(self.nodeToJSON(it))
                    inds.append(it)
        
        # Get node and link information.
        nodes, aS = self._jsonGraph(nodes)
        links = [x for x in self._jsonEdges(nodes)]
        out   = json.dumps(
            {"nodes":nodes,"links":links,"graphid":self.name,
            "scored":aS},sort_keys=True,indent=4,separators=(',',': '))
        return out
    
    
    def getCompleteLandscape(self):
        
        ''' Returns the landscape as a JSON string. 
        
        :return: a JSON string
        
        '''
        
        return self.getJSON()

    def getJSON(self):

        ''' Returns the landscape as a JSON string.
        
        :return: a JSON string
        
        '''

        # Get node and link information.
        nodes, aS = self._jsonGraph()
        nodes = [node for node in nodes if node['ml'] != None]
        links = [x for x in self._jsonEdges(nodes)]
        
        # Output to JSON string.
        out = json.dumps({"nodes":nodes,"links":links,"graphid":self.name,
                         "scored":aS},sort_keys=True,indent=4,separators=
                         (',',': '))
        return out
        


''' Newick string parsing and object interaction. A Newick string can represent
a phylogenetic tree. '''

# Date:   Feb 16 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from rearrangement import topology
from random import shuffle
from math import factorial as fact
import base

# Exception Handling

class ParsingError(Exception):
    def __init__(self,val): self.value = val
    def __str__(self): return repr(self.value)

# Class Definitions

class node(base.treeNode):
    
    ''' Node for a tree parsed from a Newick string. '''
        
    def __init__(self,lbl='',children=None,parent=None):
        
        ''' Initialize a node in a tree parsed from a Newick string.
        
        :param lbl: a label for this node
        :type lbl: a string
        :param children: an optional set of children (branches); default none
        :type children: a list of :class:`.branch` objects
        :param parent: an optional parent branch for this node; default none
        :type parent: a :class:`.branch` object
        
        '''
        
        super(node,self).__init__(lbl,children,parent)
        
    def __str__(self):
        sT, sl = self.children, self.label
        if len(self.children) > 0:
            # Perform sorting to ensure consistent naming scheme.
            l = sorted([st for st in sT],key=lambda d: d.child.label)
            return '(%s)%s' % (','.join(map(str,l)),sl)
        else: return self.label

class branch(base.treeBranch):
    
    ''' Branch for a tree parsed from a Newick string. '''
    
    def __init__(self,chi,l,parent=None,s=None):
        
        ''' Initialize a branch in a tree parsed from a Newick string. 
        
        :param chi: a child node
        :type chi: a :class:`.node` object
        :param l: a branch length
        :type l: a floating point value
        :param parent: an optional parent node; default none
        :type parent: a :class:`.node` object
        
        '''
        
        self.parent         = parent
        self.child          = chi
        self.branch_length  = l
        self.branch_support = s
        
    def __str__(self):
        st, sl = self.child, self.branch_length
        if sl > 0: return '%s:%s' % (str(st),str(sl)) # Branch Length
        else:      return '%s'    % (str(st))         # No Branch Length

# Traversal Functions

def assignParents(top):
    
    ''' Should be a one-time use function. Goes through
    and assigns parents to the parsed newick tree structure
    nodes and branches to allow for up-traversal.
    
    :param top: a top-level node for a tree (root node)
    :type top: a :class:`.node` object
    
    '''
    
    # Is a node?
    if type(top) != node: return False
    
    # Traverse.
    for br in top.children:
        br.parent       = top
        br.child.parent = br
        assignParents(br.child)
    return True

def removeBranchLengths(top):
    
    ''' Goes through and removes any stored branch lengths.
    
    :param top: a top-level node for a tree (root node)
    :type top: a :class:`.node` object
    
    '''

    # Is a node?
    if type(top) != node: return False
    
    # Traverse.
    for br in top.children:
        br.branch_length = 0.0
        removeBranchLengths(br.child)
    return True

def removeUnaryInternalNodes(top):
    
    ''' Goes through and ensures any degree-2 internal 
    nodes are smoothed into a single degree-3 internal 
    node. 
    
    :param top: a top-level node for a tree (root node)
    :type top: a :class:`.node` object    
        
    '''
    
    next_item = None
    if (len(top.children) == 1):
        pa             = top.parent
        ch             = top.children[0]
        if (pa and ch):
            st, end = pa.parent, ch.child
            st.children.remove(pa)
            br = branch(end,pa.branch_length+ch.branch_length,st)
            st.children.append(br)
            end.parent = br
            next_item  = end
    
    if not next_item:
        for b in top.children:
            removeUnaryInternalNodes(b.child)
    else: removeUnaryInternalNodes(next_item)
    
def invertAlongPathToNode(target,top):
    
    ''' DANGEROUS: Reverses all directionality to a given
    node from a top-level node. Intended as a low-level 
    function for rerooting a tree.
    
    :param target: a target node
    :type target: a :class:`.node` object
    :param top: a top-level node for a tree (root node)
    :type top: a :class:`.node` object
    
    '''
        
    def invertDirections(n,newparent=None):
        b = n.parent
        if b:
            b.child  = b.parent
            b.parent = n
            if len(n.children) != 0:
                n.children.append(b)
        n.parent = newparent        
        
    # Are nodes?
    if type(target) != node: return False
    elif type(top)  != node: return False
    
    if target == top:
        # End condition.
        invertDirections(target)
        return True
    
    for b in top.children:
        endpoint = b.child
        if invertAlongPathToNode(target,endpoint):
            invertDirections(top,b)
            top.children.remove(b)
            return True
    return False

def shuffleLeaves(top):
    
    ''' DANGEROUS: Given a top-level node, shuffle all 
    leaves in this tree.
    
    :param top: a top-level node for a tree (root node)
    :type top: a :class:`.node` object
    
    '''

    leaves = base.treeStructure.leaves(top)
    names  = [x.label for x in leaves]
    shuffle(names)
    for l in xrange(len(leaves)):
        leaves[l].label = names[l]

def getAllBranches(br):
    
    ''' Given a branch, traverse subtree and return 
    comprising branches as a list.
    
    :param br: a branch from a tree
    :type br: a :class:`.branch` object
    
    '''
    
    # Is a branch?
    if type(br) != branch: return list()
    
    # Make list.
    li = list()
    
    # Traverse.
    s = br.child
    li.extend(s.children)
    for bra in s.children: li.extend(getAllBranches(bra))
    return li
        
def isSibling(br,other):
    
    ''' Given a branch, determine if that branch is 
    adjacent to another branch. 
    
    :param br: a branch from a tree
    :type br: a :class:`.branch` object
    :param other: another branch from a tree
    :type other: a :class:`.branch` object
    
    '''

    if type(br)     != branch: return False
    if type(other)  != branch: return False
    return (br.parent == other.parent)

# Parsing Assets

class newickParser:
    
    ''' Parsing object for Newick strings. '''
    
    def __init__(self, newick):
        
        ''' Initialize this parser (with a Newick string). 
        
        :param newick: a Newick string
        :type newick: a string
        
        '''
                
        if newick == None or type(newick) != str:
            raise ValueError('Input <%s> is not a valid string!' %
                             (repr(newick)))
        self.newick           = newick
        self.parsed_structure = None
        
    def parse(self):
        
        ''' Parse the stored newick string into a topological structure.
        
        :return: the top-level root :class:`.node` object
        
        '''
        
        # Get the newick string.
        read = self.newick
        
        # Check if proper newick format.
        if read[-1] != ';':
            raise ParsingError('Input newick string does not end with ";".')
        
        length   = len(read)
        toplevel = parseNewick(read,0,length-1,None)
        assignParents(toplevel)
        self.parsed_structure = toplevel
        return toplevel
    
    def __str__(self):
        
        if not self.parsed_structure: self.parse()
        return str(self.parsed_structure) + ';'

def getBalancingBracket(newick,i):
    
    ''' Given a position of an opening bracket in a 
    newick string, i, output the closing bracket's
    position that corresponds to this opening bracket.
    
    :param newick: a Newick string
    :type newick: a string
    :param i: a position in the string (index)
    :type i: an integer < length of the string
    :return: an integer
    
    '''
    
    opencount = 0
    while (i+1 < len(newick)):
        if newick[i+1] == '(': opencount += 1 
        elif newick[i+1] == ')':
            opencount -= 1
            if (opencount < 0):
                return i+1
        i += 1
    raise ParsingError('Could not find balancing bracket.')

def getBranchLength(newick,i):
    
    ''' Given a position of a colon symbol (indicating a 
    branch length), return the branch length.
    
    :param newick: a Newick string
    :type newick: a string
    :param i: a position in the string (index)
    :type i: an integer < length of the string
    :return: an integer
    
    '''
    
    runningstr = ''
    while (i+1 < len(newick)):
        if not newick[i+1].isdigit() and not newick[i+1] == '.':
            break
        else: runningstr += newick[i+1]
        i += 1
    try: out = float(runningstr)
    except:
        raise ParsingError('Could not parse branch length as float at %d in %s.'
                           % (i,newick))
    return (i,out)

def getLeafName(newick,i):
    
    ''' Given the position of a leaf, find its complete name.
    
    :param newick: a Newick string
    :type newick: a string
    :param i: a position in the string (index)
    :type i: an integer < length of the string
    :return: an integer
    
    '''
    
    runningstr = ''
    while (i < len(newick)):
        if newick[i] in [',',':',';',')'] \
           or i == len(newick):
            return (i-1,runningstr)
        runningstr += newick[i]
        i += 1
    return (i-1,runningstr)
       
def parseNewick(newick,i,j,top):
    
    ''' Parse a newick string into a topological newick
    structure given a top-level node.
    
    :param newick: a Newick string
    :type newick: a string
    :param i: a starting position to start parsing
    :type i: an integer
    :param j: an end position to stop parsing
    :type j: an integer
    :param top: a top-level node; start parsing with None
    :type top: a :class:`.node` object
    
    '''
    
    # Input is substring [i,j] and top-level node.
    first = None
    while (i <= j):
        
        # Create new node.
        newnode = node()
        
        # Construct branch edge.
        if top != None:
            li = top.children
            nb = branch(newnode,0.0,top)
            li.append(nb)
        elif not first: first = newnode
        
        # Hit a subtree?
        if newick[i] == '(':
            k, _ = (getBalancingBracket(newick,i)), None
            brlength = 0.0
            if newick[k+1] == ':': # Has branch length?
                _, brlength = getBranchLength(newick,k+1)
            elif newick[k+1] != ',': # Has name or statistical support?
                _, name       = getLeafName(newick,k+1)
                newnode.label = name
                if newick[_+1] == ':': # Branch length after this?
                    _, brlength = getBranchLength(newick,_+1)
            if brlength > 0.0 and top: nb.branch_length = brlength
            parseNewick(newick,i+1,k,newnode)
            if (_ != None): k = _
        
        # Hit a leaf?
        else:
            k, name = getLeafName(newick,i)
            brlength = 0.0
            if (k+1 <= j):
                if newick[k+1] == ':':
                    k, brlength = getBranchLength(newick,k+1)
            if brlength > 0.0 and top: nb.branch_length = brlength
            newnode.label = name
            
        # Advance forward.
        i = k + 1 # Next symbol.
        if (i<j):
            while (i+1 <= j and newick[i] != ','): i += 1
        i += 1

    return first


''' Object model defining a sequence alignment (DNA, RNA, protein sequences). 
Handle input biological sequence alignment files for the purposes of 
phylogenetic inference. Will read all types of alignment files by utilizing the 
P4 python phylogenetic library. '''

# Date:   Jan 24 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# Imports

import os
import p4
import model
import tree
import base
from newick import newickParser
from executable import fasttree
from tempfile import NamedTemporaryFile as NTempFile
from shutil import copyfile

# Set up.

p4.var.warnReadNoFile = False

# Class Definitions


class alignment(object):
    
    ''' Wrap a biological sequence alignment to enable functionality
    necessary for phylogenetic inference. Makes use of temporary
    files; requires to be closed once no longer needed. '''
    
    def __init__(self,inal=None):
        
        ''' Instantiate an object intended to wrap an alignment for the
        purposes of running phylogenetic inference. 
        
        :param inal: An alignment file path (most formats are accepted). 
        
        '''
        
        # Intergrate the p4 phylogenetic library.
        if inal is None:
            self.data = p4.Alignment()
            
        else:
            p4.read(inal) # Read the alignment file/string.
            self.data  = p4.var.alignments[-1]

        # Augment alignment with a discrete state model.
        self.model = model.DiscreteStateModel(self)

        # Keep track of temporary files.
        self.paths = {} 

    # Internals

    def __getitem__(self,i):
        
        return self.data.sequences[i]
    
    def __str__(self):
        
        return '\n'.join(self.toStrList())
    
    def __len__(self):
        
        return len(self.data)
    
    def __iter__(self):
        
        for i in xrange(self.getNumSeqs()): yield self[i]

    def _makeFASTAFile(self):
        
        ''' PRIVATE: Make a temporary FASTA file. '''
        
        if ('fasta' in self.paths):
            return
        fh = NTempFile(delete=False)
        self.paths['fasta'] = fh.name
        self.data.writeFasta(fh.name)
        fh.close()

    def close(self):
        
        ''' Forcefully delete all temporary files and clear data. '''
        
        self.data  = None
        self.model = None
        if len(self.paths) > 0:
            for f in self.paths.values(): os.unlink(f)
            self.paths = {}

    # Accessors

    def toStrList(self):

        ''' Get all sequences as a list of strings. 

        :return: a list of strings
        
        '''

        return [self.getSequenceString(i) for i in xrange(self.getNumSeqs())]

    def getDataType(self):
        
        ''' Get the data type associated with this alignment (e.g., protein). 
        
        :return: a string indicating the data type ('protein', 'DNA')
        
        '''

        return self.data.dataType

    def getStateModel(self):

        ''' Get the state model associated with this alignment. See model
        module for more information.
        
        :return: a :class:`.model.DiscreteStateModel` object
        
        '''

        return self.model

    def getSize(self):

        ''' Return the size of the alignment, or how many characters
        there are in each respective item in the alignment. 
        
        :return: an integer
        
        '''

        return len(self)

    def getNumSeqs(self):

        ''' Return the number of sequences that are present in the 
        sequence alignment. 
        
        :return: an integer
        
        '''

        return len(self.data.sequences)

    def getDim(self):

        ''' Return the dimensionality of the sequence alignment (how 
        many different types of characters). 
        
        :return: an integer
        
        '''

        return self.data.dim

    def getSequenceString(self,i):

        ''' Acquire the ith sequence as a string. 
        
        :param i: an index in the alignment (associated with a sequence)
        :type i: an integer
        :return: a string associated with the sequence
        
        '''

        return self[i].sequence

    def getFASTA(self):

        ''' Get (and create if not already) a path to a temporary 
        FASTA file. This will be deleted upon closure of the alignment
        instance. 
        
        :return: a string associated with a path in the file system
        
        '''

        self._makeFASTAFile()
        return self.paths['fasta']

    def getApproxMLNewick(self):

        ''' Get a tree in newick format via use of FastTree that serves as
        an approximation of the maximum likelihood tree for this data. 
        
        :return: a Newick or New Hampshire string
        
        '''

        ft = fasttree(
            self.getFASTA(), isProtein=(self.data.dataType=='protein'))
        return ft.run()

    def getApproxMLTree(self):

        ''' Get a tree object for an approximation of the maximum likelihood
        tree for this data using FastTree. 
        
        :return: a :class:`.tree.tree` object
        
        '''

        return tree.tree(self.getApproxMLNewick(),check=True)

    def getTaxa(self):

        ''' Get a list of taxa names associated with the alignment. 
        
        :return: a list of strings
        
        '''

        return self.data.taxNames


class phylipFriendlyAlignment(alignment):

    ''' An alignment object that renames all comprising taxa in order
    to be able to be written as a strict Phylip file. '''

    def __init__(self,inal=None):

        # Call superclass constructor.
        super(phylipFriendlyAlignment,self).__init__(inal)

        # Keep track of all taxa names.
        self.namedict = {}

        # Acquire a temporary file handle for Nexus with proper names.
        self._makeProperNexusFile()

        # Reassign all taxa names to shorter ones (for Phylip).
        self._reassignNames()  

    def _makePhylipFile(self):

        ''' PRIVATE: Make a temporary Phylip file. '''

        if ('phylip' in self.paths): return
        fh = NTempFile(delete=False)
        self.paths['phylip'] = fh.name
        self.data.writePhylip(
            fh.name,whitespaceSeparatesNames=False)
        fh.close()
        
    def _makeProperNexusFile(self):

        ''' PRIVATE: Make a temporary Nexus file. '''

        if ('nexus' in self.paths): return
        fh = NTempFile(delete=False)
        self.paths['nexus'] = fh.name
        self.data.writeNexus(
            fName=fh.name)    
        fh.close()

    def _reassignNames(self):

        ''' PRIVATE: Reassign all sequence names to shorter 
        Phylip-friendly integer-based names. '''

        d  = self.namedict
        for item in self.data.sequences:
            _         = item.name
            ind       = self.data.sequences.index(item)
            name      = 'T' + str(ind)
            d[name]   = _
            item.name = name
            self.data.taxNames[ind] = name

    def getPhylip(self):

        ''' Get (and create if not already) a path to a temporary 
        Phylip file. This will be deleted upon closure of the alignment
        instance. 

        :return: a string associated with a path in the file system

        '''

        self._makePhylipFile()
        return self.paths['phylip']

    def writeProperNexus(self,wri):

        ''' Write a Nexus file with proper names. 
        
        :param wri: a path to a (existent or unexistent) file to write to
        :type wri: a string
        
        '''

        copyfile(self.paths['nexus'],wri)
        return True

    def reassignFromReinterpretedNewick(self,tr):

        ''' Return a Newick string with taxa names replaced with shortened forms
        as they are defined in this object. 
        
        :param tr: a Newick string
        :type tr: a string
        :return: a Newick string with all replaced names
        
        '''

        p = newickParser(tr).parse()
        nodes = base.treeStructure(p).getAllNodes()
        for node in nodes:
            l = node.label
            if l in self.namedict.values():
                for n in self.namedict:
                    if self.namedict[n] == l:
                        node.label = n
                        break
        return str(p) + ';'

    def reinterpretNewick(self,tr):

        ''' Revert the replacing of taxa names with shortened names by changing
        them back to their original form.
        
        :param tr: a Newick string
        :type tr: a string
        :return: a Newick string with all replaced names
        
        '''

        p = newickParser(tr).parse()
        nodes = base.treeStructure(p).getAllNodes()
        for node in nodes:
            l = node.label
            if l in self.namedict: node.label = self.namedict[l]
        return str(p) + ';'

    def getProperName(self,n):

        ''' Return the actual name for an integer-based sequence name 
        that was reassigned at initialization. 
        
        :param n: a shortened taxon name from this object
        :type n: a string
        :return: a string (replaced with the original taxon name)
        
        '''

        if n in self.namedict: return self.namedict[n]
        else: return None

    def getTaxa(self):

        ''' Return current taxa names in the alignment.
        
        :return: a list of shortened taxa names
        
        '''

        return [x for x in self.namedict]

    def recreateObject(self):

        ''' Reintializes the object. '''

        return phylipFriendlyAlignment(self.getFASTA())



''' An installation script for Pylogeny. '''

# Date:   Oct 16 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from setuptools import setup, Extension as extension
import os

# Metadata

from pylogeny.__version__ import VERSION
DESCRIP = 'A code framework for phylogenetic tree reconstruction, rearrangement, scoring, and for the manipulation, heuristic search, and analysis of the phylogenetic tree combinatorial space.'
LONG    = 'A Python library and code framework for phylogenetic tree reconstruction, rearrangement, scoring, and for the manipulation and analysis of the phylogenetic tree combinatorial space. Also possesses features to execute popular heuristic programs such as FastTree and RAxML to acquire approximate ML trees.'
URL     = 'http://www.github.com/AlexSafatli/Pylogeny'
AUTHOR  = 'Alex Safatli'
EMAIL   = 'safatli@cs.dal.ca'
DEPNDS  = ['networkx','pandas','mysql-python']
LINKS   = ['http://p4-phylogenetics.googlecode.com/archive/4491de464e68fdb49c7a11e06737cd34a98143ec.tar.gz']
PKGDATA = {'pylogeny':['fitch.cpp','libpllWrapper.c']}
FITCHCC = os.path.join('pylogeny','fitch.cpp')
PLLC    = os.path.join('pylogeny','libpllWrapper.c')

# Compilation for C/C++ Extensions (Fitch, Pylibpll)

pllExtension    = extension('libpllWrapper',sources=[PLLC],include_dirs=['/usr/local/include'],libraries=['pll-sse3'],library_dirs=['/usr/local/lib'])
fitchExtension = extension('fitch',sources=[FITCHCC],include_dirs=['/usr/local/include'],language="c++",extra_compile_args=['-std=c++11'])

# Setup

setup(name='pylogeny',version=VERSION,description=DESCRIP,long_description=LONG,url=URL,author=AUTHOR,author_email=EMAIL,license='MIT',packages=['pylogeny'],package_data=PKGDATA,ext_modules=[pllExtension,fitchExtension],install_requires=DEPNDS,dependency_links=LINKS,zip_safe=False)



''' Serialize a phylogenetic landscape into an SQLlite database file made up of
three components: all tree IDs and respective scores, the alignment file as a
set of sequences, and a representation of the graph as an edge list. '''

# Date:   Apr 9 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from __version__ import VERSION
from alignment import phylipFriendlyAlignment as alignment
from landscape import landscape
from database import SQLiteDatabase
import os, tree, sys

class landscapeWriter(object):

    ''' Encapsulate the writing of a landscape to a file format. '''

    def __init__(self, landscape, name):
        
        ''' Instantiates this writer.
        
        :param landscape: a landscape object
        :type landscape: a :class:`.landscape.landscape` object
        :param name: the name of this landscape
        :type name: a string
        
        '''
        
        # Fields
        self.landscape = landscape
        self.name      = name
        self.graph     = landscape.graph
        self.database  = None
        self.cleankey  = ''
        self.cleansuff = ''
        
        # Clean the input landscape name.
        self._cleanName()

    def _cleanName(self):

        if (type(self.name) != str):
            raise TypeError('Landscape name given not a string.')
        for x in self.name:
            # Replace all non-alphanumeric characters with a "_".
            if x.isalnum(): self.cleankey += x
            else: self.cleankey += '_'
        self.cleansuff = '%s.landscape' % (self.cleankey)

    def _schema(self):
        
        # Get the database object.
        dbobj = self.database
        
        # Create all content tables.
        dbobj.newTable('alignment',('seqid','integer'),('seqname','text'),
                       ('sequence','text'))
        dbobj.newTable('trees',('treeid','integer'),('name','text'),
                       ('newick','text'),('origin','text'),('ml','real'),
                       ('pars','real'),('explored','boolean'),
                       ('structure','text'))
        dbobj.newTable('graph',('source','integer'),('origin','integer'))
        dbobj.newTable('locks',('treeid','integer'),('branchid','integer'))
        
        # Create a table to store metadata.
        dbobj.newTable('metadata',('key','text'),('value','text'))

    def _dump(self,path='.'):

        # Open the file.
        fpath = os.path.join(path,self.cleansuff) # Generate file path.
        if os.path.isfile(fpath):
            os.unlink(fpath) # Remove file if already exists.
        o = SQLiteDatabase(fpath) # Open an SQLite database at that location.
        self.database = o # Assign this object to field.
        
        # Construct the schema for the landscape.
        self._schema()
        
        # Include metadata about the landscape. Required for parsing.
        o.insertRecord('metadata',['name',self.name])
        o.insertRecord('metadata',['version',VERSION])
        
        # Add the alignment as a set of sequence records labelled by taxa.
        index = 0
        if self.landscape.alignment != None:
            for s in self.landscape.alignment:
                o.insertRecord('alignment',[index,s.name,s.sequence])
                index += 1
        
        # Add all of the trees (most memory intensive elements in the DB).
        for i in self.landscape.iterNodes():
            t = self.landscape.getTree(i)
            s = t.getStructure()
            find = self.landscape.findTreeTopologyByStructure(s)
            if (find == None):
                raise AssertionError('Tree topology search failed (%s).' % (
                    str(i)))
            elif (find == i): # Verify unique identity of this tree.
                # Insert this tree as a record.
                name = t.getName()
                newi = t.getNewick()
                ori = t.getOrigin()
                scs = [t.getScore()[x] for x in xrange(0,len(t.getScore()))]
                o.insertRecord('trees',[i,name,newi,ori,scs[0],scs[1],
                                        self.landscape.getNode(i)['explored'],
                                        s])
            else:
                sys.stderr.write(
                    'Warning: Tree %s has identical structure to %s.\n' % (
                    str(find),str(i)))
        
        # Add the graph by copying its adjacency list.
        adj_list = self.graph.edges_iter()
        for tupl in adj_list:
            source,target = [int(x) for x in tupl]
            o.insertRecord('graph',[source,target])
        
        # Add all of the locks.
        for l in self.landscape.getLocks():
            tree = l.topology.toTree()
            treeIndex = self.landscape.indexOf(tree)
            o.insertRecord('locks',[treeIndex,l.getBranchIndex()])
        
        # Close the file.
        o.close()

        # Return the path.
        return fpath

    def writeFile(self,path='.'):
        
        ''' Write the landscape serialized file to given path.
        
        :param path: a directory path, defaulting to the current one
        :type path: a string
        :return: the relative filepath to the written file
        
        '''

        return self._dump(path)

class landscapeParser(object):

    ''' Encapsulates the construction of a landscape
    object from a sqlite landscape file. '''

    def __init__(self,path):
        
        ''' Instantiate this parser. 
        
        :param path: the filepath to the landscape file
        :type path: a string
        
        '''
        
        self.file = path
        self.metadata = None
        self.name = None
        self.database = None
        self.alignment = None
        self.treemap = {}
        self.landscape = None

    def getName(self):
        
        ''' Acquire the name of the parsed landscape.
        
        :return: a string
        
        '''
        
        if self.name is None and self.metadata != None:
                for metadata in self.metadata:
                    if metadata[0] == 'name': self.name = metadata[1]
        elif self.name is None: self.name = os.path.splitext(
            os.path.basename(self.file))[0]
        return self.name

    def _getLandscapeMetadata(self):
        
        ''' Check metadata in landscape if present. '''
        
        tables = self.database.getTables()
        if 'metadata' in tables:
            return self.database.getRecords('metadata')
        else: raise IOError('Incompatible landscape DB format.')

    def _makeAlignment(self):
        
        ''' Construct a pseudo-FASTA string to create a new alignment from. '''
        
        pseudofasta = ''
        sequences = self.database.iterRecords('alignment')
        for sequence in sequences:
            _,name,seq = sequence
            if name != '' and seq != '':
                pseudofasta += '>%s\n%s\n' % (name,seq)
        if pseudofasta != '':
            self.alignment = alignment(str(pseudofasta))
        
    def _getTrees(self):
    
        floatIfNotNone = lambda d: float(d) if d != None else d
        intIfNotNone   = lambda d: int(d) if d != None else d
        for t in self.database.iterRecords('trees'):
            struct = None
            if len(t) == 7:
                treeid,name,newick,orig,ml,pars,exp = t
            else:
                treeid,name,newick,orig,ml,pars,exp,struct = t
            if newick != '':
                if (struct == None):
                    i = self.landscape.addTreeByNewick(str(
                        newick),score=False,check=False)
                else:
                    i = self.landscape.addTreeByNewick(str(
                        newick),score=False,check=False,struct=struct)
                trobj = self.landscape.getTree(i)
                trobj.setOrigin(str(orig))
                trobj.setName(name)
                trobj.setScore([floatIfNotNone(ml),intIfNotNone(pars)])
                self.treemap[int(treeid)] = trobj
                self.landscape.getVertex(i).setExplored(bool(exp))

    def _getGraph(self):
        
        getIDForTree = lambda d: self.landscape.findTreeTopologyByStructure(
            d.getStructure())
        for e in self.database.iterRecords('graph'):
            raw_s,raw_t = e
            source,target = getIDForTree(self.treemap[raw_s]),getIDForTree(
                self.treemap[raw_t])
            if source is None or target is None:
                raise IOError(
                    'Adjacency list record has source or target as None.')
            self.landscape.graph.add_edge(source,target)

    def _applyLocks(self):

        for l in self.database.iterRecords('locks'):
            self.landscape.lockBranchFoundInTreeByIndex(self.treemap[l[0]],l[1])
        
    def parse(self):
        
        ''' Parse the file.
        
        :return: a tuple of a :class:`.landscape.landscape` object and its name (a string)
        
        '''

        # Get database object.
        fpath = self.file
        if not os.path.isfile(fpath): raise IOError('Landscape file not found.')
        o = SQLiteDatabase(fpath)
        self.database = o
        
        # Get metadata.
        self.metadata = self._getLandscapeMetadata()
        
        # Get the alignment.
        self._makeAlignment()
        
        # Create an empty landscape structure.
        self.landscape = landscape(self.alignment,root=False)     
        
        # Add all of the trees.
        self._getTrees()
        
        # Set landscape operator.
        if len(self.landscape) > 0:
            index = self.landscape.getNodeNames()[-1]
            self.landscape.setOperator(self.landscape.getTree(
                index).getOrigin())
        
        # Ensure all edges are established.
        self._getGraph()
        
        # Apply locks.
        self._applyLocks()

        return (self.landscape,self.getName())
